/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comte
 * (Université de Franche-Comté), France.
 *
 * Copyright © Université de Franche-Comté 2007-2012.
 *
 * Corresponding author: Matteo Cypriani <mcy@lm7.fr>
 *
 ***********************************************************************
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL:
 * http://www.cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's authors, the holder
 * of the economic rights, and the successive licensors have only
 * limited liability.
 *
 * In this respect, the user's attention is drawn to the risks
 * associated with loading, using, modifying and/or developing or
 * reproducing the software by the user in light of its specific status
 * of free software, that may mean that it is complicated to manipulate,
 * and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users
 * are therefore encouraged to load and test the software's suitability
 * as regards their requirements in conditions enabling the security of
 * their systems and/or data to be ensured and, more generally, to use
 * and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 ***********************************************************************
 */


#include "outputtcpsocketevaal.hh"
#include "request.hh"
#include "resultlist.hh"
#include "area.hh"
#include "stock.hh"
#include "posexcept.hh"

#include <iostream>
#include <sstream>
#include <cstdio> // For perror()
#include <boost/lexical_cast.hpp>

using namespace std ;



/* *** Constructors *** */


OutputTCPSocketEvAAL::
OutputTCPSocketEvAAL(const string &_remote_host,
                     const uint_fast16_t _remote_port):
  OutputNetworkSocket(_remote_host, _remote_port)
{
  if (! init_socket())
    throw error_opening_output_file("TCP socket (EvAAL)") ;
}



/* *** Operations *** */


/**
 * @return \em true if the socket were successfully opened.
 * @return \em false in case of error.
 */
bool OutputTCPSocketEvAAL::init_socket()
{
  sockfd = socket(AF_INET, SOCK_STREAM, 0) ;
  if (sockfd < 0)
    {
      perror("TCP socket creation failed") ;
      return false ;
    }

  struct sockaddr_in server_description ;
  memset(&server_description, 0, sizeof(server_description)) ;
  server_description.sin_family = AF_INET ;
  // Server IP address:
  server_description.sin_addr.s_addr = inet_addr(remote_host.c_str()) ;
  // Listening port on the server:
  server_description.sin_port = htons(remote_port) ;

  int ret = connect(sockfd, (struct sockaddr *)&server_description,
                    sizeof(server_description)) ;
  if (ret < 0)
    {
      perror("Cannot bind the TCP socket") ;
      return false ;
    }

  return true ;
}


/**
 * @param results Must contain only one element, since the EvAAL format
 * accepts only one algorithm result.
 */
void OutputTCPSocketEvAAL::write(const ResultList &results)
{
  assert(results.get_results().size() <= 1) ;
  if (! results.get_results().empty())
    write(results.get_results()[0]) ;
}


void OutputTCPSocketEvAAL::write(const Result &result)
{
  const Point3D &position = result.get_position() ;
  Timestamp request_time = result.get_request()->get_time_sent() ;
  int area_of_interest = area_of_interest_number(position) ;

  ostringstream str ;
  str
      << "OwlPS "
      << position.get_x() << ' '
      << position.get_y() << ' '
      << static_cast<uint64_t>(request_time) << ' ' ;

  if (area_of_interest > 0)
    str << area_of_interest ;

  str << '\n' ;

  if (send_data(str.str()))
    acknowledge() ;
}


int OutputTCPSocketEvAAL::
area_of_interest_number(const Point3D &position) const
{
  const Area *const area = Stock::in_which_area_is(position) ;
  if (area == NULL)
    return 0 ;

  int aoi_number = 0 ;
  try
    {
      aoi_number = boost::lexical_cast<int>(area->get_name()) ;
    }
  catch (boost::bad_lexical_cast &e)
    {
      cerr
          << "Cannot convert the area of interest's name (\""
          << area->get_name() << "\") into an integer value: "
          << e.what() << '\n' ;
    }

  return aoi_number ;
}


/**
 * Sends the text buffer 'data'.
 */
bool OutputTCPSocketEvAAL::send_data(const string &data) const
{
  // Do *not* add one extra character for the '\0', it must not be sent!
  unsigned int data_len = data.size() ;

  ssize_t nsent = send(sockfd, data.c_str(), data_len, 0) ;
  if (nsent != static_cast<ssize_t>(data_len))
    {
      perror("Error sending result data through the TCPEvAAL socket") ;
      return false ;
    }
  return true ;
}


bool OutputTCPSocketEvAAL::acknowledge() const
{
  char ack_buf[10] ;
  ssize_t nrecv = recv(sockfd, ack_buf, 10, 0) ;
  if (nrecv < 0)
    perror("Error receiving acknowledgement") ;

  ack_buf[9] = '\0' ;
  string ack(ack_buf) ;
  if (ack == "Published")
    return true ;

  if (ack == "NotParsed")
    {
      cerr << "TCP server error: string not parsed.\n" ;
      return false ;
    }

  cerr
      << "TCP server error: unknown acknoledgment string \""
      << ack << "\".\n" ;
  return false ;
}
