/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comte
 * (Université de Franche-Comté), France.
 *
 * Copyright © Université de Franche-Comté 2007-2012.
 *
 * Corresponding author: Matteo Cypriani <mcy@lm7.fr>
 *
 ***********************************************************************
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL:
 * http://www.cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's authors, the holder
 * of the economic rights, and the successive licensors have only
 * limited liability.
 *
 * In this respect, the user's attention is drawn to the risks
 * associated with loading, using, modifying and/or developing or
 * reproducing the software by the user in light of its specific status
 * of free software, that may mean that it is complicated to manipulate,
 * and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users
 * are therefore encouraged to load and test the software's suitability
 * as regards their requirements in conditions enabling the security of
 * their systems and/or data to be ensured and, more generally, to use
 * and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 ***********************************************************************
 */


#include "result.hh"
#include "request.hh"
#include "area.hh"
#include "stock.hh"
#include "configuration.hh"

#include <sstream>

using namespace std ;



/* *** Constructors *** */


Result::Result(const Request *_request,
               const std::string &_algorithm,
               const Point3D &_position,
               const Point3D &real_position):
  request(_request), algorithm(_algorithm), position(_position)
{
  compute_error(real_position) ;
}



/* *** Read accessors *** */


string Result::in_which_area() const
{
  const Area *area = Stock::in_which_area_is(position) ;
  if (area == NULL)
    return "" ;

  return area->get_name() ;
}



/* *** Write accessors *** */


void Result::compute_error(const Point3D &real_position)
{
  if (Configuration::bool_value("output.2d-error"))
    error = position.distance_2d(real_position) ;
  else
    error = position.distance(real_position) ;
}



/* *** Operators *** */


Result& Result::operator=(const Result &source)
{
  if (this == &source)
    return *this ;

  request = source.request ;
  algorithm = source.algorithm ;
  position = source.position ;
  error = source.error ;

  return *this ;
}


bool Result::operator==(const Result &source) const
{
  return
    request == source.request &&
    algorithm == source.algorithm &&
    position == source.position &&
    error == source.error ;
}



/**
 * This function creates a CSV string from the Result's data.
 *
 * The format used is the following:
 *   Algorithm_name;X;Y;Z;Error;Area_name
 *
 * @return The result as a CSV string, \em without a trailing '\\n'.
 */
const string Result::to_csv() const
{
  ostringstream csv_line ;

  csv_line
      << algorithm
      << ';' << position.get_x()
      << ';' << position.get_y()
      << ';' << position.get_z()
      << ';' << error
      << ';' ;

  const Area *const area = Stock::in_which_area_is(position) ;
  if (area != NULL)
    csv_line << area->get_name() ;

  return csv_line.str() ;
}



std::ostream& operator<<(ostream &os, const Result &r)
{
  os
      << "The result of the algorithm " << r.algorithm
      << " is: " << r.position
      << " (area: \"" << r.in_which_area() << "\")" ;
  if (r.error >= 0)
    os << ", error = " << r.error << " m" ;
  return os ;
}
