/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comte
 * (Université de Franche-Comté), France.
 *
 * Copyright © Université de Franche-Comté 2007-2012.
 *
 * Corresponding author: Matteo Cypriani <mcy@lm7.fr>
 *
 ***********************************************************************
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL:
 * http://www.cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's authors, the holder
 * of the economic rights, and the successive licensors have only
 * limited liability.
 *
 * In this respect, the user's attention is drawn to the risks
 * associated with loading, using, modifying and/or developing or
 * reproducing the software by the user in light of its specific status
 * of free software, that may mean that it is complicated to manipulate,
 * and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users
 * are therefore encouraged to load and test the software's suitability
 * as regards their requirements in conditions enabling the security of
 * their systems and/or data to be ensured and, more generally, to use
 * and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 ***********************************************************************
 */


#ifndef _OWLPS_POSITIONING_RESULT_HH_
#define _OWLPS_POSITIONING_RESULT_HH_

class Request ;

#include "point3d.hh"

#include <string>

/// Represents a result computed by a positioning algorithm
class Result
{
protected:
  /// Request sent by the mobile
  const Request *request ;
  /// Algorithm used to compute the position
  std::string algorithm ;
  /// Computed coordinates of the mobile
  Point3D position ;
  /// Distance error between the real coordinates and the computed point
  /** The error is set to -1 if the real coordinates are unknown. */
  float error ;

public:
  Result(const Request *_request = NULL,
         const std::string &_algorithm = "UnknownAlgorithm"):
    request(_request), algorithm(_algorithm), error(-1) {}
  Result(const Request *_request, const std::string &_algorithm,
         const Point3D &_position):
    request(_request), algorithm(_algorithm), position(_position),
    error(-1) {}
  Result(const Request *_request, const std::string &_algorithm,
         const Point3D &_position, const Point3D &real_position) ;
  ~Result(void) {}

  /** @name Read accessors */
  //@{
  const Request* get_request(void) const ;
  const std::string& get_algorithm(void) const ;
  const Point3D& get_position(void) const ;
  float get_error(void) const ;
  std::string in_which_area(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  void compute_error(const Point3D &real_position) ;
  //@}

  /** @name Operators */
  //@{
  Result& operator=(const Result &source) ;
  bool operator==(const Result &source) const ;
  bool operator!=(const Result &source) const ;
  //@}

  /** @name Conversion accessors */
  //@{
  /// Converts to a CSV string
  const std::string to_csv(void) const ;
  //@}

  /// Displays a Result
  friend std::ostream& operator<<(std::ostream &os, const Result &r) ;
} ;



/* *** Read accessors *** */


inline const Request* Result::get_request() const
{
  return request ;
}


inline const std::string& Result::get_algorithm() const
{
  return algorithm ;
}


inline const Point3D& Result::get_position() const
{
  return position ;
}


inline float Result::get_error() const
{
  return error ;
}



/* *** Operators *** */


inline bool Result::operator!=(const Result &source) const
{
  return !(*this == source) ;
}



#endif // _OWLPS_POSITIONING_RESULT_HH_
