/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comte
 * (Université de Franche-Comté), France.
 *
 * Copyright © Université de Franche-Comté 2007-2012.
 *
 * Corresponding author: Matteo Cypriani <mcy@lm7.fr>
 *
 ***********************************************************************
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL:
 * http://www.cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's authors, the holder
 * of the economic rights, and the successive licensors have only
 * limited liability.
 *
 * In this respect, the user's attention is drawn to the risks
 * associated with loading, using, modifying and/or developing or
 * reproducing the software by the user in light of its specific status
 * of free software, that may mean that it is complicated to manipulate,
 * and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users
 * are therefore encouraged to load and test the software's suitability
 * as regards their requirements in conditions enabling the security of
 * their systems and/or data to be ensured and, more generally, to use
 * and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 ***********************************************************************
 */


#include "textfilereader.hh"
#include "posexcept.hh"

#include <iostream>

using namespace std ;



/* *** Constructors *** */


/**
 * @param _file_name The name of the file to open.
 * @throw error_opening_input_file if the file cannot be opened.
 */
TextFileReader::TextFileReader(const string &_file_name):
  file_name(_file_name), current_line_nb(0)
{
  if (file_name == "-")
    {
      file_buf = file.rdbuf() ;
      file.std::ios::rdbuf(std::cin.rdbuf()) ;
    }
  else
    {
      file.open(file_name.c_str()) ;
      if (! file)
        throw error_opening_input_file(file_name) ;
    }
}


TextFileReader::~TextFileReader()
{
  if (file_name == "-")
    file.std::ios::rdbuf(file_buf) ;
  else
    file.close() ;
}



/* *** Operations *** */


/**
 * Blank lines or lines containing only blank characters are skipped.
 * Tabs and spaces at the begining of a line are deleted.
 * If the first non-blank character of a line is a #, then the line is
 * skipped (note that comments at the end of a line are *not* handled).
 * @param text Output argument; unchanged in case of error.
 * @return \em false in case of error, \em true else.
 */
bool TextFileReader::read_nonblank_line(string &text)
{
  string line ;
  string::size_type first_non_blank = 0 ;

  do
    if (! read_line(line))
      return false ;
  while ((first_non_blank = line.find_first_not_of(" \t"))
         == string::npos || line.at(first_non_blank) == '#') ;

  line.erase(0, first_non_blank) ;
  text = line ;

  return true ;
}


/**
 * @return The string read, or an empty string in case of error.
 */
bool TextFileReader::read_line(string &text)
{
  if (eof_close())
    return false ;

  getline(file, text) ;
  ++current_line_nb ;
  return true ;
}


/**
 * Tests if #file is opened and there is something more to
 * read. If the end of file is reached, the stream is closed.
 * @return \em true if the end of file is reached or if the file is
 * already closed.
 * @return \em false if there is something more to read.
 */
bool TextFileReader::eof_close()
{
  if (file)
    return false ;

  if (file.eof())
    {
      file.close() ;
      return true ;
    }

  return false ;
}
