/*
 * This file is part of the Owl Positioning System (OwlPS).
 * OwlPS is a project of the University of Franche-Comte
 * (Université de Franche-Comté), France.
 *
 * Copyright © Université de Franche-Comté 2007-2012.
 *
 * Corresponding author: Matteo Cypriani <mcy@lm7.fr>
 *
 ***********************************************************************
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL:
 * http://www.cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's authors, the holder
 * of the economic rights, and the successive licensors have only
 * limited liability.
 *
 * In this respect, the user's attention is drawn to the risks
 * associated with loading, using, modifying and/or developing or
 * reproducing the software by the user in light of its specific status
 * of free software, that may mean that it is complicated to manipulate,
 * and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users
 * are therefore encouraged to load and test the software's suitability
 * as regards their requirements in conditions enabling the security of
 * their systems and/or data to be ensured and, more generally, to use
 * and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 ***********************************************************************
 */


#include "waypoint.hh"
#include "building.hh"

#include <sstream>

using namespace std ;
using std::tr1::unordered_set ;



/* *** Constructors *** */


/**
 * @param _b A pointer to the (first) Building to add to #buildings.
 * If it is NULL, #buildings will remain empty.
 * @param _x X coordinate.
 * @param _y Y coordinate.
 * @param _z Z coordinate.
 */
Waypoint::Waypoint(const Building *_b,
                   const float &_x, const float &_y, const float &_z
                  ): Point3D(_x, _y, _z)
{
  if (_b != NULL)
    buildings.insert(const_cast<Building*>(_b)) ;
}


/**
 * @param _b A pointer to the (first) Building to add to #buildings.
 * If it is NULL, #buildings will remain empty.
 * @param p Coordinates of the Waypoint.
 */
Waypoint::Waypoint(const Building *_b, const Point3D &p): Point3D(p)
{
  if (_b != NULL)
    buildings.insert(const_cast<Building*>(_b)) ;
}


/**
 * Clears #buildings, but does not deallocate the values pointed by
 * the elements into it.
 */
Waypoint::~Waypoint()
{
  buildings.clear() ;
}



/* *** Operators *** */


Waypoint& Waypoint::operator=(const Waypoint &wp)
{
  if (this == &wp)
    return *this ;

  this->Point3D::operator=(wp) ;
  buildings = wp.buildings ;

  return *this ;
}


bool Waypoint::operator==(const Waypoint &wp) const
{
  if (this == &wp)
    return true ;

  return
    this->Point3D::operator==(wp) &&
    buildings == wp.buildings ;
}


Waypoint::operator std::string() const
{
  ostringstream csv ;
  csv << (Point3D) *this ;

  for (unordered_set<Building*>::const_iterator i = buildings.begin() ;
       i != buildings.end() ; ++i)
    csv << ';' << (*i)->get_name() ;

  return csv.str() ;
}



ostream &operator<<(ostream &os, const Waypoint &wp)
{
  // Coordinates
  os << (Point3D) wp ;

  // List of buildings
  if (wp.buildings.empty())
    os << '\n' << "Belongs to no building!" ;
  else
    for (unordered_set<Building*>::const_iterator i =
           wp.buildings.begin() ; i != wp.buildings.end() ; ++i)
      os << '\n' << **i ;

  return os ;
}
